# -*- coding: utf-8 -*-


import matplotlib.pyplot as plt
import random
import numpy as np
import math
import copy
import numpy.typing as npt


#Funktionen
#Setzt alle Folgewerte auf 0, sobald Teilchen Hoehenschicht verlaesst
def set_zeros(n_in_layer: npt.NDArray, upper_bound: float, lower_bound: float) -> npt.NDArray:

    for i in range(np.shape(n_in_layer)[0]):
        for j in range(np.shape(n_in_layer)[1]):

            #Prueft, ob sich der Wert an Index [i, j] innerhalb des definierten, geschlossenen Intervalls befindet
            if(not lower_bound <= n_in_layer[i, j] <= upper_bound):

                #Ersetzt die Reihe mit einer neuen Liste aus Teilliste bis zum ersten Null-Index und einer mit Nullen gefuellten
                #Liste als Padding, um den Shape des Original-Arrays zu erhalten. Danach bricht der Innere Loop ab, da das Pruefen
                #nachfolgender Werte nicht mehr erfordrlich ist.

                n_in_layer[i] = np.concatenate((n_in_layer[i, 0:j-1], np.zeros(len(n_in_layer[i])-j+1)))
                break

    return n_in_layer

#Zaehlt N zu Zeitpunkt T in Hoehenschicht
def count_n_t(n_in_layer: npt.NDArray) -> np.array:

    #Erzeugt ein neues numpy-Array mittels List Comprehension. Dabei wird mit (np.logical_not(n_in_layer == 0)).view(np.int8)
    #zuerst eine Binary-Repraesentation angefertigt. Alle mit Werten besetzen Felder werden durch 1 ersetzt. Felder mit Wert 0 behalten
    #diesen bei. '.T' gibt eine transponierte Ansicht zurueck. Anschließend werden die Einsen in den Columns nur noch summiert.

    return np.array([np.sum(col) for col in (np.logical_not(n_in_layer == 0)).view(np.int8).T])

#Errechnet die Teilchen, die die Schicht pro Zeitschritt verlassen haben
def get_diff(n_t: npt.NDArray) -> npt.NDArray:

    #Erzeugt ein neues, um eine Position nach Rechts verschobenes Numpy-Array aus n_t.
    #Anschließend wird der Wert auf Index 0 auf 1500 gesetzt. Dies ist notwendig, da der Wert auf Index 0 nach dem Shift dem
    #Wert entsprechen wuerde, der vorher auf Index[-1] gesetzt war. Zum Zeitpunkt 0 koennen noch keine Teilchen die Schicht verlassen haben,
    #daher entspricht hier der abzuziehende Wert dem Startwert.
     
    shift_right = np.roll(n_t, 1)
    np.put(shift_right, 0, 1500)

    diff = np.subtract(shift_right, n_t)

    return diff

#Ermittelt die durchschnittlichen Zeitschritte, die sich ein Teilchen in der Höhenschicht befindet
def get_avg_t_in_layer(n_in_layer: npt.NDArray) -> npt.NDArray:
    t_in_layer = np.zeros(np.shape(n_in_layer)[0])

    for i in range(np.shape(n_in_layer)[0]):
        for j in range(np.shape(n_in_layer)[1]):
            if(n_in_layer[i, j] == 0):
                t_in_layer[i] = j
                break
    
    return np.average(t_in_layer)

#Wertet Hoehenschicht aus
def evaluate_layer(trace_x: npt.NDArray, trace_y: npt.NDArray, trace_z: npt.NDArray, lower_bound: float, upper_bound: float) -> tuple:

    #Hier werden lediglich die einzelnen Funktionen aufgerufen.

    n_in_layer = set_zeros(copy.deepcopy(trace_z), upper_bound, lower_bound)
    n_t = count_n_t(n_in_layer)
    avg_t_in_layer = get_avg_t_in_layer(n_in_layer)
    diff = get_diff(n_t)
    return (n_t, avg_t_in_layer, diff)


#Parameter Random Walk
N = 1000 #Anzahl der Beobachtungen
T = 1500# Dauer der Beobachtungen
schrittlaenge = 3
dt = 1 

# Parameter Ebenen / Schicht
d = 10  # Schichtdicke
hoehe = 50 # Mitte der Beobachtungsschicht = Startpunkt der Teilchen 
boden = 0 # Lage des Bodens auf Z-Achse
gr_unten = hoehe-1/2*d  # untere Schichtgrenze
gr_oben = hoehe+1/2*d   # obere Schichtgrenze

# Parameter Einsteins Formel
R = 8.31e+7   #Gaskonstante (Produkt von Avogadro- und Boltzmann-Konstante)
temp = 290    #Absolute Temperatur
Na = 6e+23    #Avogadro-Konstante
Eta = .0135   #Viskosität (Zähigkeit) des Lösungmittels
r = .5e-4 	  #Radius des Teilchens in cm
rm = 10000*r  #Radius in Mikron

# Startkoordiaten
x = [0]*N
y = [0]*N
z = [hoehe]*N 

# numpy arrays für die Speicherung der Koordinaten auf den einzelnen Achsen
# Werte für jedes Teilchen über die gesamte Zeit d.h. shape = (N,T)
shape = (N,T)
trace_x = np.zeros(shape)
trace_y = np.zeros(shape)
trace_z = np.zeros(shape)
trace_z[:,0] = hoehe # z startet bei allen Teilchen auf definierter Höhe

#Abstände
trace_abstand = np.zeros(shape) # Abstände aller einzelnen Teilchen des Random Walk zu allen Zeitpunkten
abstand_einstein = [0]*T # quadratischer Abstand nach Einsteins Formel
qmd = [0]*T #quadratischer mittler Abstand aller Teilchen beim Random Walk 
differenz = [0]*T # Differenz von berechnetem und simuliertem quadratischen Abstand

#Länge ununterbrochener Aufenthalt (la) in der Schicht + Marker für Verlassen (la[n,1]=1) der Schicht
shape = (N, 2)
la = np.zeros(shape) 
la[:,0] = 1 # Minimaler Aufenthalt ist 1 Zeitschritt
avg_la = 0 
zaehlen=[0]*T 


 
for t in range(1,T,dt):
    for n in range(N):
        Delta_y = schrittlaenge*(random.random()-0.5)
        Delta_x = schrittlaenge*(random.random()-0.5)
        Delta_z = schrittlaenge*(random.random()-0.5)   
        

        x[n] = x[n] + Delta_x
        trace_x[n][t]=x[n]
        
        y[n] = y[n] + Delta_y
        trace_y[n][t]=y[n]
        

        z[n] = z[n] + Delta_z
        
        # Boden Reflexion mit Radius (effektive untere Grenze: Bodenhöhe + Radius)
        if z[n]-rm < boden: 
            z[n] = abs(boden-z[n])+rm
            
        trace_z[n][t]=z[n]
        
        # Überprüfung Aufenthalt in der Schicht
        if trace_z[n][t]< hoehe + 1/2*d + rm and trace_z[n][t]> hoehe -(1/2*d) -rm and la[n][1] == 0:
            la[n][0] += 1
        else:
            la[n][1] = 1
            
        # Abstand der einzelnen Teilchen in drei Dimensionen (Raumdiagonale)
        trace_abstand[n][t]=(x[n]**2+y[n]**2+(z[n]-hoehe)**2)**0.5
     
    # Quadratischer mittlerer Abstand
    qmd[t]= (sum(trace_abstand[:,t])/N)**2
    
    #Als Referenz wird der Abstand nach Einstein berechnet und gespeichert.
    # die Formel wurde für die dreidimensionale Berechnung mit Wurzel 3 multipliziert
    abstand_einstein[t]=((R*temp/Na/math.pi/Eta/r)**0.5*t**0.5*10000)**2
    
    #Zur Kontrolle wird die Differenz aus den Abständen des Random-Walks und n. Einstein berechnet.
    differenz[t]=qmd[t] - abstand_einstein[t]
    
        
avg_la = sum(la[:,0])/N
       
      

print
print()
print("Teilchenanzahl: ", N)
print("Ausserhalb der Beobachtungsschicht: ", int(sum(la[:,1])), )
print("Durchgehend beobachtbar: ", int(N-sum(la[:,1])), )
print("Dauer Minium: ",la[:,0].min())        
print("Dauer Maxium: ",la[:,0].max())
n_t, avg_t_in_layer, diff = evaluate_layer(trace_x, trace_y, trace_z, lower_bound=hoehe-1/2*d, upper_bound=hoehe+1/2*d)
t = np.arange(np.shape(n_t)[0])
print("Dauer Durchschnitt: ", round(avg_la, 2) )


# Darstellung mittlere quadratische Abweichung
fig, (ax1, ax2) = plt.subplots(figsize=(12,5), nrows=1, ncols=2)
ax1.set_title('MSD Random Walk')
ax1.set_xlabel('τ')
ax1.set_ylabel('<Δ²>')
ax1.plot(abstand_einstein, 'r', label=('Referenz Einstein'))
ax1.plot(qmd, 'black', label='<Δ²>')
ax1.legend(loc=0)
ax2.set_title('Differenz: <Δ²> - Einstein')
ax2.plot(differenz)

plt.show()


# Darstellung Random walk mit Boden (grün) und Schichtgrenzen (schwarz)
fig, ax = plt.subplots(figsize=(20,10))

for n in range(N):
   
    plt.plot(trace_x[n],  trace_z[n]) #trace_y[n],

rand = ax.get_xlim() 
ax.set_ylim(boden-5, gr_oben+5)
ax.set_title('Seitenansicht Beobachtungsschicht')
ax.set_xlabel('x-Wert')
ax.set_ylabel('Höhe')
line1, = ax.plot([rand[0],rand[1]],[hoehe, hoehe], color="black", label="Höhe", linestyle='--')
line2, = ax.plot([rand[0],rand[1]],[gr_oben, gr_oben],color="black", label="oben", linewidth=4)
line3, = ax.plot([rand[0],rand[1]],[gr_unten, gr_unten], color="black", label="unten", linewidth=4)
line4, = ax.plot([rand[0],rand[1]],[boden, boden], color="green", label="Boden", linewidth=4)
plt.show()


# Grafik Statistik
fig, ax = plt.subplots()
fig.set_size_inches(20,9)
ax.bar(t, n_t)
ax.bar(t, diff)

plt.show()

